/*
   Here are three implementations of
   the longest common subsequence
   dynamic programming problem.
      lcs(i,j) = 1+lcs(i-1,j-1),                if a1[i]==a2[j]
                 max{ lcs(i-1,j), lcs(i,j-1) }  if a1[i]!=a2[j]
   with
      lcs(0,j) = 0
      lcs(i,0) = 0
*/

//#define Max(A,B) ( ((A) > (B)) ? A : B )  // BAD macro!!
int Max(int A, int B){ return (A > B) ? A : B; }

/*
   Compute the longest common subsequence, lcs(a1,a2,i,j) where
   i is the length of array a1 and j is the length of array a2.

   Here is a purely recursive (and very inefficient,
   exponential) implementation of lcs(i,j).
*/
int lcs1(int* a1, int* a2, int i, int j)
{
   int length;
   if ( i == 0 || j == 0 )
      length = 0;
   else if (a1[i-1] == a2[j-1] )
      length = 1 + lcs1(a1, a2, i-1, j-1);
   else
      length = Max( lcs1(a1, a2, i-1, j), lcs1(a1, a2, i, j-1) );

   return length;
}



/*
   Here is a "top down dynamic programming"
   implementation of lcs(i,j).
   This is also called "memoization dynamic programming".
*/
int lcs2Helper(int* a1, int* a2, int i, int j, int n, int m, int* dpTable);

int lcs2(int* a1, int* a2, int i, int j)
{
   int length, k;

   int* dpTable = malloc(i*j*sizeof(int));
   if (dpTable == NULL)  // error check
   {
      fprintf(stderr, "lcs2: not enough memory for array of size %d.\n", j);
      return -1;
   }

   for (k = 0; k < i*j; k++)
      dpTable[k] = -1;  // initialize the table with "null" values

   length = lcs2Helper(a1, a2, i, j, i, j, dpTable);

   // print the dynamic programming table
   if (i <= 30 && j <= 30)
   {  int row, col;
      printf("Here is the top-down, lcs dynamic programming table.\n");
      for (row = 0; row < i; row++)
      {
         for(col = 0; col < j; col++)
         {
            printf(" %*d", 2, dpTable[col+row*j]);
         }
         printf("\n");
      }
   }

   free(dpTable);
   return length;
}

int lcs2Helper(int* a1, int* a2, int i, int j, int n, int m, int* dpTable)
{
   int length;
   if ( i == 0 || j == 0 )            // Compare this with the purely recursive verion.
      length = 0;                     // This is kind of a "short circuited" recursion.
   else if (dpTable[(j-1)+(i-1)*m]>=0)// If the answer is already in the table,
      length = dpTable[(j-1)+(i-1)*m];// then use that value.
   else                               // Otherwise, do the recursion.
   {
      if (a1[i-1] == a2[j-1] )
         length = 1 + lcs2Helper(a1, a2, i-1, j-1, n, m, dpTable);
      else
         length = Max( lcs2Helper(a1, a2, i-1, j,   n, m, dpTable),
                       lcs2Helper(a1, a2, i,   j-1, n, m, dpTable) );

      dpTable[(j-1)+(i-1)*m] = length;  // Cache this result in the table.
   }
   return length;
}



/*
   Here is a "bottom up dynamic programming"
   implementation of lcs(i,j).
*/
int lcs3(int* a1, int* a2, int i, int j)
{
   int length, row, col;

   int* dpTable = malloc((i+1)*(j+1)*sizeof(int));
   if (dpTable == NULL)  // error check
   {
      fprintf(stderr, "lcs3: not enough memory for array of size %d.\n", j);
      return -1;
   }

   for (col = 0; col <= j; col++)  // initialize top row
      dpTable[col+0*(j+1)] = 0;

   for (row = 0; row <= i; row++)  // initialize left most column
      dpTable[0+row*(j+1) ] = 0;

   for (row = 1; row <= i; row++)
      for (col = 1; col <= j; col++)
      {
         if (a1[row-1] == a2[col-1] )
            dpTable[col+row*(j+1)] = 1 + dpTable[(col-1)+(row-1)*(j+1)];
         else
            dpTable[col+row*(j+1)] = Max( dpTable[(col-1)+ row   *(j+1)],
                                          dpTable[ col   +(row-1)*(j+1)] );
      }

   // print the dynamic programming table
   if (i <= 30 && j <= 30)
   {  int row, col;
      printf("Here is the bottom-up, lcs dynamic programming table.\n");
      for (row = 0; row <= i; row++)
      {
         for(col = 0; col <= j; col++)
         {
            printf(" %*d", 2, dpTable[col+row*(j+1)]);
         }
         printf("\n");
      }
   }

   length = dpTable[j+i*(j+1)];
   free(dpTable);
   return length;
}
